@@ -3,39 +3,64 @@
 import tkinter
 from tkinter import TclError
 from types import *
+from Tkconstants import *
+
 CallableTypes = (FunctionType, MethodType,
-		 BuiltinFunctionType, BuiltinMethodType,
-		 ClassType, InstanceType)
+		 BuiltinFunctionType, BuiltinMethodType)
+
+TkVersion = eval(tkinter.TK_VERSION)
+TclVersion = eval(tkinter.TCL_VERSION)
 
-# Older versions of tkinter don't define these variables
-try:
-	TkVersion = eval(tkinter.TK_VERSION)
-except AttributeError:
-	TkVersion = 3.6
-try:
-	TclVersion = eval(tkinter.TCL_VERSION)
-except AttributeError:
-	TclVersion = 7.3
 
 def _flatten(tuple):
 	res = ()
 	for item in tuple:
 		if type(item) in (TupleType, ListType):
 			res = res + _flatten(item)
-		else:
+		elif item is not None:
 			res = res + (item,)
 	return res
 
 def _cnfmerge(cnfs):
-	if type(cnfs) in (NoneType, DictionaryType, StringType):
+	if type(cnfs) is DictionaryType:
+		_fixgeometry(cnfs)
+		return cnfs
+	elif type(cnfs) in (NoneType, StringType):
+		
 		return cnfs
 	else:
 		cnf = {}
 		for c in _flatten(cnfs):
+			_fixgeometry(c)
 			for k, v in c.items():
 				cnf[k] = v
 		return cnf
 
+if TkVersion >= 4.0:
+	_fixg_warning = "Warning: patched up pre-Tk-4.0 geometry option\n"
+	def _fixgeometry(c):
+		if c and c.has_key('geometry'):
+			wh = _parsegeometry(c['geometry'])
+			if wh:
+				# Print warning message -- once
+				global _fixg_warning
+				if _fixg_warning:
+					import sys
+					sys.stderr.write(_fixg_warning)
+					_fixg_warning = None
+				w, h = wh
+				c['width'] = w
+				c['height'] = h
+				del c['geometry']
+	def _parsegeometry(s):
+		from string import splitfields
+		fields = splitfields(s, 'x')
+		if len(fields) == 2:
+			return tuple(fields)
+		# else: return None
+else:
+	def _fixgeometry(c): pass
+
 class Event:
 	pass
 
@@ -95,8 +120,8 @@ class BooleanVar(Variable):
 	def get(self):
 		return self._tk.getboolean(self._tk.getvar(self._name))
 
-def mainloop():
-	_default_root.tk.mainloop()
+def mainloop(n=0):
+	_default_root.tk.mainloop(n)
 
 def getint(s):
 	return _default_root.tk.getint(s)
@@ -150,6 +175,7 @@ class Misc:
 		return self._nametowidget(name)
 	def after(self, ms, func=None, *args):
 		if not func:
+			# I'd rather use time.sleep(ms*0.001)
 			self.tk.call('after', ms)
 		else:
 			# XXX Disgusting hack to clean up after calling func
@@ -161,7 +187,16 @@ class Misc:
 					tk.deletecommand(tmp[0])
 			name = self._register(callit)
 			tmp.append(name)
-			self.tk.call('after', ms, name)
+			return self.tk.call('after', ms, name)
+	def after_idle(self, func, *args):
+		return apply(self.after, ('idle', func) + args)
+	def after_cancel(self, id):
+		self.tk.call('after', 'cancel', id)
+	def bell(self, displayof=None):
+		if displayof:
+			self.tk.call('bell', '-displayof', displayof)
+		else:
+			self.tk.call('bell', '-displayof', self._w)
 	# XXX grab current w/o window argument
 	def grab_current(self):
 		name = self.tk.call('grab', 'current', self._w)
@@ -180,7 +215,7 @@ class Misc:
 	def lower(self, belowThis=None):
 		self.tk.call('lower', self._w, belowThis)
 	def option_add(self, pattern, value, priority = None):
-		self.tk.call('option', 'add', pattern, priority)
+		self.tk.call('option', 'add', pattern, value, priority)
 	def option_clear(self):
 		self.tk.call('option', 'clear')
 	def option_get(self, name, className):
@@ -322,25 +357,38 @@ class Misc:
 		self.tk.call('update')
 	def update_idletasks(self):
 		self.tk.call('update', 'idletasks')
+	def bind(self, sequence, func=None, add=''):
+		if add: add = '+'
+		if func:
+		    name = self._register(func, self._substitute)
+		    self.tk.call('bind', self._w, sequence, 
+				 (add + name,) + self._subst_format)
+		else:
+		    return self.tk.call('bind', self._w, sequence)
 	def unbind(self, sequence):
 		self.tk.call('bind', self._w, sequence, '')
-	def bind(self, sequence, func, add=''):
+	def bind_all(self, sequence, func=None, add=''):
 		if add: add = '+'
-		name = self._register(func, self._substitute)
-		self.tk.call('bind', self._w, sequence, 
-			     (add + name,) + self._subst_format)
-	def bind_all(self, sequence, func, add=''):
+		if func:
+		    name = self._register(func, self._substitute)
+		    self.tk.call('bind', 'all' , sequence, 
+				 (add + name,) + self._subst_format)
+		else:
+		    return self.tk.call('bind', 'all', sequence)
+	def unbind_all(self, sequence):
+		self.tk.call('bind', 'all' , sequence, '')
+	def bind_class(self, className, sequence, func=None, add=''):
 		if add: add = '+'
-		name = self._register(func, self._substitute)
-		self.tk.call('bind', 'all' , sequence, 
-			     (add + name,) + self._subst_format)
-	def bind_class(self, className, sequence, func, add=''):
-		if add: add = '+'
-		name = self._register(func, self._substitute)
-		self.tk.call('bind', className , sequence, 
-			     (add + name,) + self._subst_format)
-	def mainloop(self):
-		self.tk.mainloop()
+		if func:
+		    name = self._register(func, self._substitute)
+		    self.tk.call('bind', className , sequence, 
+				 (add + name,) + self._subst_format)
+		else:
+		    return self.tk.call('bind', className, sequence)
+	def unbind_class(self, className, sequence):
+		self.tk.call('bind', className , sequence, '')
+	def mainloop(self, n=0):
+		self.tk.mainloop(n)
 	def quit(self):
 		self.tk.quit()
 	def _getints(self, string):
@@ -352,10 +400,14 @@ class Misc:
 	def _getboolean(self, string):
 		if string:
 			return self.tk.getboolean(string)
-	def _options(self, cnf):
-		cnf = _cnfmerge(cnf)
+	def _options(self, cnf, kw = None):
+		if kw:
+			cnf = _cnfmerge((cnf, kw))
+		else:
+			cnf = _cnfmerge(cnf)
 		res = ()
 		for k, v in cnf.items():
+			if k[-1] == '_': k = k[:-1]
 			if type(v) in CallableTypes:
 				v = self._register(v)
 			res = res + ('-'+k, v)
@@ -376,13 +428,13 @@ class Misc:
 			name = tail
 		return w
 	def _register(self, func, subst=None):
-		f = _CallSafely(func, subst).__call__
+		f = self._wrap(func, subst)
 		name = `id(f)`
 		if hasattr(func, 'im_func'):
 			func = func.im_func
-		if hasattr(func, 'func_name') and \
-		   type(func.func_name) == type(''):
-			name = name + func.func_name
+		if hasattr(func, '__name__') and \
+		   type(func.__name__) == type(''):
+			name = name + func.__name__
 		self.tk.createcommand(name, f)
 		return name
 	register = _register
@@ -420,25 +472,28 @@ class Misc:
 		e.x_root = tk.getint(X)
 		e.y_root = tk.getint(Y)
 		return (e,)
+	def _report_exception(self):
+		import sys
+		exc, val, tb = sys.exc_type, sys.exc_value, sys.exc_traceback
+		root = self._root()
+		root.report_callback_exception(exc, val, tb)
+	def _wrap(self, func, subst=None):
+		return CallWrapper(func, subst, self).__call__
 
-class _CallSafely:
-	def __init__(self, func, subst=None):
+class CallWrapper:
+	def __init__(self, func, subst, widget):
 		self.func = func
 		self.subst = subst
+		self.widget = widget
 	def __call__(self, *args):
-		if self.subst:
-			args = self.apply_func(self.subst, args)
-		args = self.apply_func(self.func, args)
-	def apply_func(self, func, args):
-		import sys
 		try:
-			return apply(func, args)
+			if self.subst:
+				args = apply(self.subst, args)
+			return apply(self.func, args)
 		except SystemExit, msg:
 			raise SystemExit, msg
 		except:
-			import traceback
-			print "Exception in Tkinter callback"
-			traceback.print_exc()
+			self.widget._report_exception()
 
 class Wm:
 	def aspect(self, 
@@ -513,6 +568,7 @@ class Wm:
 class Tk(Misc, Wm):
 	_w = '.'
 	def __init__(self, screenName=None, baseName=None, className='Tk'):
+		global _default_root
 		self.master = None
 		self.children = {}
 		if baseName is None:
@@ -520,16 +576,34 @@ class Tk(Misc, Wm):
 			baseName = os.path.basename(sys.argv[0])
 			if baseName[-3:] == '.py': baseName = baseName[:-3]
 		self.tk = tkinter.create(screenName, baseName, className)
+		# Version sanity checks
+		tk_version = self.tk.getvar('tk_version')
+		if tk_version != tkinter.TK_VERSION:
+		    raise RuntimeError, \
+		    "tk.h version (%s) doesn't match libtk.a version (%s)" \
+		    % (tkinter.TK_VERSION, tk_version)
+		tcl_version = self.tk.getvar('tcl_version')
+		if tcl_version != tkinter.TCL_VERSION:
+		    raise RuntimeError, \
+		    "tcl.h version (%s) doesn't match libtcl.a version (%s)" \
+		    % (tkinter.TCL_VERSION, tcl_version)
+		if TkVersion < 4.0:
+		    raise RuntimeError, \
+			  "Tk 4.0 or higher is required; found Tk %s" \
+			  % str(TkVersion)
 		self.tk.createcommand('tkerror', _tkerror)
 		self.tk.createcommand('exit', _exit)
 		self.readprofile(baseName, className)
+		if not _default_root:
+			_default_root = self
 	def destroy(self):
 		for c in self.children.values(): c.destroy()
 		self.tk.call('destroy', self._w)
 	def __str__(self):
 		return self._w
 	def readprofile(self, baseName, className):
-		import os
+		##print __import__
+		import os, pdb
 		if os.environ.has_key('HOME'): home = os.environ['HOME']
 		else: home = os.curdir
 		class_tcl = os.path.join(home, '.%s.tcl' % className)
@@ -537,6 +611,7 @@ class Tk(Misc, Wm):
 		base_tcl = os.path.join(home, '.%s.tcl' % baseName)
 		base_py = os.path.join(home, '.%s.py' % baseName)
 		dir = {'self': self}
+		##pdb.run('from Tkinter import *', dir)
 		exec 'from Tkinter import *' in dir
 		if os.path.isfile(class_tcl):
 			print 'source', `class_tcl`
@@ -550,12 +625,16 @@ class Tk(Misc, Wm):
 		if os.path.isfile(base_py):
 			print 'execfile', `base_py`
 			execfile(base_py, dir)
+	def report_callback_exception(self, exc, val, tb):
+		import traceback
+		print "Exception in Tkinter callback"
+		traceback.print_exception(exc, val, tb)
 
 class Pack:
-	def config(self, cnf={}):
+	def config(self, cnf={}, **kw):
 		apply(self.tk.call, 
 		      ('pack', 'configure', self._w) 
-		      + self._options(cnf))
+		      + self._options(cnf, kw))
 	pack = config
 	def __setitem__(self, key, value):
 		Pack.config({key: value})
@@ -575,7 +654,7 @@ class Pack:
 	info = newinfo
 	_noarg_ = ['_noarg_']
 	def propagate(self, flag=_noarg_):
-		if boolean is Pack._noarg_:
+		if flag is Pack._noarg_:
 			return self._getboolean(self.tk.call(
 				'pack', 'propagate', self._w))
 		else:
@@ -586,10 +665,10 @@ class Pack:
 				   self.tk.call('pack', 'slaves', self._w)))
 
 class Place:
-	def config(self, cnf={}):
+	def config(self, cnf={}, **kw):
 		apply(self.tk.call, 
 		      ('place', 'configure', self._w) 
-		      + self._options(cnf))
+		      + self._options(cnf, kw))
 	place = config
 	def __setitem__(self, key, value):
 		Place.config({key: value})
@@ -628,15 +707,20 @@ class Widget(Misc, Pack, Place):
 		if self.master.children.has_key(self._name):
 			self.master.children[self._name].destroy()
 		self.master.children[self._name] = self
-	def __init__(self, master, widgetName, cnf={}, extra=()):
-		cnf = _cnfmerge(cnf)
-		Widget._setup(self, master, cnf)
+	def __init__(self, master, widgetName, cnf={}, kw={}, extra=()):
+		if kw:
+			cnf = _cnfmerge((cnf, kw))
 		self.widgetName = widgetName
-		apply(self.tk.call, (widgetName, self._w) + extra)
+		Widget._setup(self, master, cnf)
+		apply(self.tk.call, (widgetName, self._w)+extra)
 		if cnf:
 			Widget.config(self, cnf)
-	def config(self, cnf=None):
-		cnf = _cnfmerge(cnf)
+	def config(self, cnf=None, **kw):
+		# XXX ought to generalize this so tag_config etc. can use it
+		if kw:
+			cnf = _cnfmerge((cnf, kw))
+		else:
+			cnf = _cnfmerge(cnf)
 		if cnf is None:
 			cnf = {}
 			for x in self.tk.split(
@@ -654,6 +738,8 @@ class Widget(Misc, Pack, Place):
 		apply(self.tk.call, (self._w, 'configure')
 		      + self._options(cnf))
 	def __getitem__(self, key):
+		if TkVersion >= 4.0:
+			return self.tk.call(self._w, 'cget', '-' + key)
 		v = self.tk.splitlist(self.tk.call(
 			self._w, 'configure', '-' + key))
 		return v[4]
@@ -670,10 +756,15 @@ class Widget(Misc, Pack, Place):
 			del self.master.children[self._name]
 		self.tk.call('destroy', self._w)
 	def _do(self, name, args=()):
-		return apply(self.tk.call, (self._w, name) + args) 
+		return apply(self.tk.call, (self._w, name) + args)
+	# XXX The following method seems out of place here
+##	def unbind_class(self, seq):
+##		Misc.unbind_class(self, self.widgetName, seq)
 
 class Toplevel(Widget, Wm):
-	def __init__(self, master=None, cnf={}):
+	def __init__(self, master=None, cnf={}, **kw):
+		if kw:
+			cnf = _cnfmerge((cnf, kw))
 		extra = ()
 		if cnf.has_key('screen'):
 			extra = ('-screen', cnf['screen'])
@@ -681,28 +772,29 @@ class Toplevel(Widget, Wm):
 		if cnf.has_key('class'):
 			extra = extra + ('-class', cnf['class'])
 			del cnf['class']
-		Widget.__init__(self, master, 'toplevel', cnf, extra)
+		Widget.__init__(self, master, 'toplevel', cnf, {}, extra)
 		root = self._root()
 		self.iconname(root.iconname())
 		self.title(root.title())
 
 class Button(Widget):
-	def __init__(self, master=None, cnf={}):
-		Widget.__init__(self, master, 'button', cnf)
-	def tk_butEnter(self, *dummy):
-		self.tk.call('tk_butEnter', self._w)
-	def tk_butLeave(self, *dummy):
-		self.tk.call('tk_butLeave', self._w)
-	def tk_butDown(self, *dummy):
-		self.tk.call('tk_butDown', self._w)
-	def tk_butUp(self, *dummy):
-		self.tk.call('tk_butUp', self._w)
+	def __init__(self, master=None, cnf={}, **kw):
+		Widget.__init__(self, master, 'button', cnf, kw)
+	def tkButtonEnter(self, *dummy):
+		self.tk.call('tkButtonEnter', self._w)
+	def tkButtonLeave(self, *dummy):
+		self.tk.call('tkButtonLeave', self._w)
+	def tkButtonDown(self, *dummy):
+		self.tk.call('tkButtonDown', self._w)
+	def tkButtonUp(self, *dummy):
+		self.tk.call('tkButtonUp', self._w)
 	def flash(self):
 		self.tk.call(self._w, 'flash')
 	def invoke(self):
 		self.tk.call(self._w, 'invoke')
 
 # Indices:
+# XXX I don't like these -- take them away
 def AtEnd():
 	return 'end'
 def AtInsert(*args):
@@ -721,8 +813,8 @@ def At(x, y=None):
 		return '@' + `x` + ',' + `y`
 
 class Canvas(Widget):
-	def __init__(self, master=None, cnf={}):
-		Widget.__init__(self, master, 'canvas', cnf)
+	def __init__(self, master=None, cnf={}, **kw):
+		Widget.__init__(self, master, 'canvas', cnf, kw)
 	def addtag(self, *args):
 		self._do('addtag', args)
 	def addtag_above(self, tagOrId):
@@ -749,14 +841,14 @@ class Canvas(Widget):
 		self.tk.call(self._w, 'bind', tagOrId, sequence, 
 			     (add + name,) + self._subst_format)
 	def canvasx(self, screenx, gridspacing=None):
-		return self.tk.getint(self.tk.call(
+		return self.tk.getdouble(self.tk.call(
 			self._w, 'canvasx', screenx, gridspacing))
 	def canvasy(self, screeny, gridspacing=None):
-		return self.tk.getint(self.tk.call(
+		return self.tk.getdouble(self.tk.call(
 			self._w, 'canvasy', screeny, gridspacing))
 	def coords(self, *args):
 		return self._do('coords', args)
-	def _create(self, itemType, args): # Args: (value, value, ..., cnf={})
+	def _create(self, itemType, args, kw): # Args: (value, value, ..., cnf={})
 		args = _flatten(args)
 		cnf = args[-1]
 		if type(cnf) in (DictionaryType, TupleType):
@@ -766,23 +858,25 @@ class Canvas(Widget):
 		return self.tk.getint(apply(
 			self.tk.call,
 			(self._w, 'create', itemType) 
-			+ args + self._options(cnf)))
-	def create_arc(self, *args):
-		return Canvas._create(self, 'arc', args)
-	def create_bitmap(self, *args):
-		return Canvas._create(self, 'bitmap', args)
-	def create_line(self, *args):
-		return Canvas._create(self, 'line', args)
-	def create_oval(self, *args):
-		return Canvas._create(self, 'oval', args)
-	def create_polygon(self, *args):
-		return Canvas._create(self, 'polygon', args)
-	def create_rectangle(self, *args):
-		return Canvas._create(self, 'rectangle', args)
-	def create_text(self, *args):
-		return Canvas._create(self, 'text', args)
-	def create_window(self, *args):
-		return Canvas._create(self, 'window', args)
+			+ args + self._options(cnf, kw)))
+	def create_arc(self, *args, **kw):
+		return self._create('arc', args, kw)
+	def create_bitmap(self, *args, **kw):
+		return self._create('bitmap', args, kw)
+	def create_image(self, *args, **kw):
+		return self._create('image', args, kw)
+	def create_line(self, *args, **kw):
+		return self._create('line', args, kw)
+	def create_oval(self, *args, **kw):
+		return self._create('oval', args, kw)
+	def create_polygon(self, *args, **kw):
+		return self._create('polygon', args, kw)
+	def create_rectangle(self, *args, **kw):
+		return self._create('rectangle', args, kw)
+	def create_text(self, *args, **kw):
+		return self._create('text', args, kw)
+	def create_window(self, *args, **kw):
+		return self._create('window', args, kw)
 	def dchars(self, *args):
 		self._do('dchars', args)
 	def delete(self, *args):
@@ -815,24 +909,25 @@ class Canvas(Widget):
 		return self.tk.getint(self._do('index', args))
 	def insert(self, *args):
 		self._do('insert', args)
-	def itemconfig(self, tagOrId, cnf=None):
-		if cnf is None:
+	def itemconfig(self, tagOrId, cnf=None, **kw):
+		if cnf is None and not kw:
 			cnf = {}
 			for x in self.tk.split(
 				self._do('itemconfigure', (tagOrId))):
 				cnf[x[0][1:]] = (x[0][1:],) + x[1:]
 			return cnf
-		if type(cnf) == StringType:
+		if type(cnf) == StringType and not kw:
 			x = self.tk.split(self._do('itemconfigure',
 						   (tagOrId, '-'+cnf,)))
 			return (x[0][1:],) + x[1:]
-		self._do('itemconfigure', (tagOrId,) + self._options(cnf))
+		self._do('itemconfigure', (tagOrId,)
+			 + self._options(cnf, kw))
 	def lower(self, *args):
 		self._do('lower', args)
 	def move(self, *args):
 		self._do('move', args)
-	def postscript(self, cnf={}):
-		return self._do('postscript', self._options(cnf))
+	def postscript(self, cnf={}, **kw):
+		return self._do('postscript', self._options(cnf, kw))
 	def tkraise(self, *args):
 		self._do('raise', args)
 	lift = tkraise
@@ -845,23 +940,23 @@ class Canvas(Widget):
 	def select_adjust(self, tagOrId, index):
 		self.tk.call(self._w, 'select', 'adjust', tagOrId, index)
 	def select_clear(self):
-		self.tk.call(self._w, 'select', 'clear')
+		self.tk.call(self._w, 'select', 'clear', 'end')
 	def select_from(self, tagOrId, index):
-		self.tk.call(self._w, 'select', 'from', tagOrId, index)
+		self.tk.call(self._w, 'select', 'set', tagOrId, index)
 	def select_item(self):
 		self.tk.call(self._w, 'select', 'item')
 	def select_to(self, tagOrId, index):
 		self.tk.call(self._w, 'select', 'to', tagOrId, index)
 	def type(self, tagOrId):
 		return self.tk.call(self._w, 'type', tagOrId) or None
-	def xview(self, index):
-		self.tk.call(self._w, 'xview', index)
-	def yview(self, index):
-		self.tk.call(self._w, 'yview', index)
+	def xview(self, *args):
+		apply(self.tk.call, (self._w, 'xview')+args)
+	def yview(self, *args):
+		apply(self.tk.call, (self._w, 'yview')+args)
 
 class Checkbutton(Widget):
-	def __init__(self, master=None, cnf={}):
-		Widget.__init__(self, master, 'checkbutton', cnf)
+	def __init__(self, master=None, cnf={}, **kw):
+		Widget.__init__(self, master, 'checkbutton', cnf, kw)
 	def deselect(self):
 		self.tk.call(self._w, 'deselect')
 	def flash(self):
@@ -874,8 +969,8 @@ class Checkbutton(Widget):
 		self.tk.call(self._w, 'toggle')
 
 class Entry(Widget):
-	def __init__(self, master=None, cnf={}):
-		Widget.__init__(self, master, 'entry', cnf)
+	def __init__(self, master=None, cnf={}, **kw):
+		Widget.__init__(self, master, 'entry', cnf, kw)
 	def tk_entryBackspace(self):
 		self.tk.call('tk_entryBackspace', self._w)
 	def tk_entryBackword(self):
@@ -900,34 +995,44 @@ class Entry(Widget):
 	def select_adjust(self, index):
 		self.tk.call(self._w, 'select', 'adjust', index)
 	def select_clear(self):
-		self.tk.call(self._w, 'select', 'clear')
+		self.tk.call(self._w, 'select', 'clear', 'end')
 	def select_from(self, index):
-		self.tk.call(self._w, 'select', 'from', index)
+	        self.tk.call(self._w, 'select', 'set', index)
+	def select_present(self):
+		return self.tk.getboolean(
+			self.tk.call(self._w, 'select', 'present'))
+	def select_range(self, start, end):
+		self.tk.call(self._w, 'select', 'range', start, end)
 	def select_to(self, index):
 		self.tk.call(self._w, 'select', 'to', index)
-	def select_view(self, index):
-		self.tk.call(self._w, 'select', 'view', index)
+	def view(self, index):
+		self.tk.call(self._w, 'view', index)
 
 class Frame(Widget):
-	def __init__(self, master=None, cnf={}):
-		cnf = _cnfmerge(cnf)
+	def __init__(self, master=None, cnf={}, **kw):
+		cnf = _cnfmerge((cnf, kw))
 		extra = ()
 		if cnf.has_key('class'):
 			extra = ('-class', cnf['class'])
 			del cnf['class']
-		Widget.__init__(self, master, 'frame', cnf, extra)
+		Widget.__init__(self, master, 'frame', cnf, {}, extra)
 	def tk_menuBar(self, *args):
 		apply(self.tk.call, ('tk_menuBar', self._w) + args)
 
 class Label(Widget):
-	def __init__(self, master=None, cnf={}):
-		Widget.__init__(self, master, 'label', cnf)
+	def __init__(self, master=None, cnf={}, **kw):
+		Widget.__init__(self, master, 'label', cnf, kw)
 
 class Listbox(Widget):
-	def __init__(self, master=None, cnf={}):
-		Widget.__init__(self, master, 'listbox', cnf)
+	def __init__(self, master=None, cnf={}, **kw):
+		Widget.__init__(self, master, 'listbox', cnf, kw)
 	def tk_listboxSingleSelect(self):
-		self.tk.call('tk_listboxSingleSelect', self._w) 
+		if TkVersion >= 4.0:
+			self['selectmode'] = 'single'
+		else:
+			self.tk.call('tk_listboxSingleSelect', self._w) 
+	def activate(self, index):
+		self.tk.call(self._w, 'activate', index)
 	def curselection(self):
 		return self.tk.splitlist(self.tk.call(
 			self._w, 'curselection'))
@@ -947,22 +1052,34 @@ class Listbox(Widget):
 		self.tk.call(self._w, 'scan', 'dragto', x, y)
 	def select_adjust(self, index):
 		self.tk.call(self._w, 'select', 'adjust', index)
-	def select_clear(self):
-		self.tk.call(self._w, 'select', 'clear')
-	def select_from(self, index):
-		self.tk.call(self._w, 'select', 'from', index)
-	def select_to(self, index):
-		self.tk.call(self._w, 'select', 'to', index)
+	if TkVersion >= 4.0:
+		def select_anchor(self, index):
+			self.tk.call(self._w, 'selection', 'anchor', index)
+		def select_clear(self, first, last=None):
+			self.tk.call(self._w,
+				     'selection', 'clear', first, last)
+		def select_includes(self, index):
+			return self.tk.getboolean(self.tk.call(
+				self._w, 'selection', 'includes', index))
+		def select_set(self, first, last=None):
+			self.tk.call(self._w, 'selection', 'set', first, last)
+	else:
+		def select_clear(self):
+			self.tk.call(self._w, 'select', 'clear')
+		def select_from(self, index):
+			self.tk.call(self._w, 'select', 'from', index)
+		def select_to(self, index):
+			self.tk.call(self._w, 'select', 'to', index)
 	def size(self):
 		return self.tk.getint(self.tk.call(self._w, 'size'))
-	def xview(self, index):
-		self.tk.call(self._w, 'xview', index)
-	def yview(self, index):
-		self.tk.call(self._w, 'yview', index)
+	def xview(self, *what):
+		apply(self.tk.call, (self._w, 'xview')+what)
+	def yview(self, *what):
+		apply(self.tk.call, (self._w, 'yview')+what)
 
 class Menu(Widget):
-	def __init__(self, master=None, cnf={}):
-		Widget.__init__(self, master, 'menu', cnf)
+	def __init__(self, master=None, cnf={}, **kw):
+		Widget.__init__(self, master, 'menu', cnf, kw)
 	def tk_bindForTraversal(self):
 		self.tk.call('tk_bindForTraversal', self._w)
 	def tk_mbPost(self):
@@ -987,24 +1104,24 @@ class Menu(Widget):
 		self.tk.call('tk_mbButtonDown', self._w)
 	def activate(self, index):
 		self.tk.call(self._w, 'activate', index)
-	def add(self, itemType, cnf={}):
+	def add(self, itemType, cnf={}, **kw):
 		apply(self.tk.call, (self._w, 'add', itemType) 
-		      + self._options(cnf))
-	def add_cascade(self, cnf={}):
-		self.add('cascade', cnf)
-	def add_checkbutton(self, cnf={}):
-		self.add('checkbutton', cnf)
-	def add_command(self, cnf={}):
-		self.add('command', cnf)
-	def add_radiobutton(self, cnf={}):
-		self.add('radiobutton', cnf)
-	def add_separator(self, cnf={}):
-		self.add('separator', cnf)
+		      + self._options(cnf, kw))
+	def add_cascade(self, cnf={}, **kw):
+		self.add('cascade', cnf or kw)
+	def add_checkbutton(self, cnf={}, **kw):
+		self.add('checkbutton', cnf or kw)
+	def add_command(self, cnf={}, **kw):
+		self.add('command', cnf or kw)
+	def add_radiobutton(self, cnf={}, **kw):
+		self.add('radiobutton', cnf or kw)
+	def add_separator(self, cnf={}, **kw):
+		self.add('separator', cnf or kw)
 	def delete(self, index1, index2=None):
 		self.tk.call(self._w, 'delete', index1, index2)
-	def entryconfig(self, index, cnf={}):
+	def entryconfig(self, index, cnf={}, **kw):
 		apply(self.tk.call, (self._w, 'entryconfigure', index)
-		      + self._options(cnf))
+		      + self._options(cnf, kw))
 	def index(self, index):
 		i = self.tk.call(self._w, 'index', index)
 		if i == 'none': return None
@@ -1020,16 +1137,16 @@ class Menu(Widget):
 			self._w, 'yposition', index))
 
 class Menubutton(Widget):
-	def __init__(self, master=None, cnf={}):
-		Widget.__init__(self, master, 'menubutton', cnf)
+	def __init__(self, master=None, cnf={}, **kw):
+		Widget.__init__(self, master, 'menubutton', cnf, kw)
 
 class Message(Widget):
-	def __init__(self, master=None, cnf={}):
-		Widget.__init__(self, master, 'message', cnf)
+	def __init__(self, master=None, cnf={}, **kw):
+		Widget.__init__(self, master, 'message', cnf, kw)
 
 class Radiobutton(Widget):
-	def __init__(self, master=None, cnf={}):
-		Widget.__init__(self, master, 'radiobutton', cnf)
+	def __init__(self, master=None, cnf={}, **kw):
+		Widget.__init__(self, master, 'radiobutton', cnf, kw)
 	def deselect(self):
 		self.tk.call(self._w, 'deselect')
 	def flash(self):
@@ -1040,25 +1157,27 @@ class Radiobutton(Widget):
 		self.tk.call(self._w, 'select')
 
 class Scale(Widget):
-	def __init__(self, master=None, cnf={}):
-		Widget.__init__(self, master, 'scale', cnf)
+	def __init__(self, master=None, cnf={}, **kw):
+		Widget.__init__(self, master, 'scale', cnf, kw)
 	def get(self):
 		return self.tk.getint(self.tk.call(self._w, 'get'))
 	def set(self, value):
 		self.tk.call(self._w, 'set', value)
 
 class Scrollbar(Widget):
-	def __init__(self, master=None, cnf={}):
-		Widget.__init__(self, master, 'scrollbar', cnf)
+	def __init__(self, master=None, cnf={}, **kw):
+		Widget.__init__(self, master, 'scrollbar', cnf, kw)
 	def get(self):
 		return self._getints(self.tk.call(self._w, 'get'))
-	def set(self, totalUnits, windowUnits, firstUnit, lastUnit):
-		self.tk.call(self._w, 'set', 
-			     totalUnits, windowUnits, firstUnit, lastUnit)
+	def set(self, *args):
+		apply(self.tk.call, (self._w, 'set')+args)
 
 class Text(Widget):
-	def __init__(self, master=None, cnf={}):
-		Widget.__init__(self, master, 'text', cnf)
+	def __init__(self, master=None, cnf={}, **kw):
+		Widget.__init__(self, master, 'text', cnf, kw)
+		self.bind('<Delete>', self.bspace)
+	def bspace(self, *args):
+	        self.delete('insert')
 	def tk_textSelectTo(self, index):
 		self.tk.call('tk_textSelectTo', self._w, index)
 	def tk_textBackspace(self):
@@ -1079,8 +1198,8 @@ class Text(Widget):
 		return self.tk.call(self._w, 'get', index1, index2)
 	def index(self, index):
 		return self.tk.call(self._w, 'index', index)
-	def insert(self, index, chars):
-		self.tk.call(self._w, 'insert', index, chars)
+	def insert(self, index, chars, *args):
+		apply(self.tk.call, (self._w, 'insert', index, chars)+args)
 	def mark_names(self):
 		return self.tk.splitlist(self.tk.call(
 			self._w, 'mark', 'names'))
@@ -1092,6 +1211,21 @@ class Text(Widget):
 		self.tk.call(self._w, 'scan', 'mark', y)
 	def scan_dragto(self, y):
 		self.tk.call(self._w, 'scan', 'dragto', y)
+	def search(self, pattern, index, stopindex=None,
+		   forwards=None, backwards=None, exact=None,
+		   regexp=None, nocase=None, count=None):
+	    args = [self._w, 'search']
+	    if forwards: args.append('-forwards')
+	    if backwards: args.append('-backwards')
+	    if exact: args.append('-exact')
+	    if regexp: args.append('-regexp')
+	    if nocase: args.append('-nocase')
+	    if count: args.append('-count'); args.append(count)
+	    if pattern[0] == '-': args.append('--')
+	    args.append(pattern)
+	    args.append(index)
+	    if stopindex: args.append(stopindex)
+	    return apply(self.tk.call, tuple(args))
 	def tag_add(self, tagName, index1, index2=None):
 		self.tk.call(
 			self._w, 'tag', 'add', tagName, index1, index2)
@@ -1103,10 +1237,10 @@ class Text(Widget):
 		self.tk.call(self._w, 'tag', 'bind', 
 			     tagName, sequence, 
 			     (add + name,) + self._subst_format)
-	def tag_config(self, tagName, cnf={}):
+	def tag_config(self, tagName, cnf={}, **kw):
 		apply(self.tk.call, 
-		      (self._w, 'tag', 'configure', tagName) 
-		      + self._options(cnf))
+		      (self._w, 'tag', 'configure', tagName)
+		      + self._options(cnf, kw))
 	def tag_delete(self, *tagNames):
 		apply(self.tk.call, (self._w, 'tag', 'delete') + tagNames)
 	def tag_lower(self, tagName, belowThis=None):
@@ -1126,27 +1260,109 @@ class Text(Widget):
 	def tag_remove(self, tagName, index1, index2=None):
 		self.tk.call(
 			self._w, 'tag', 'remove', tagName, index1, index2)
-	def yview(self, what):
-		self.tk.call(self._w, 'yview', what)
-	def yview_pickplace(self, what):
-		self.tk.call(self._w, 'yview', '-pickplace', what)
+	def window_cget(self, index, option):
+		return self.tk.call(self._w, 'window', 'cget', index, option)
+	def window_config(self, index, cnf={}, **kw):
+		apply(self.tk.call, 
+		      (self._w, 'window', 'configure', index)
+		      + self._options(cnf, kw))
+	def window_create(self, index, cnf={}, **kw):
+		apply(self.tk.call, 
+		      (self._w, 'window', 'create', index)
+		      + self._options(cnf, kw))
+	def window_names(self):
+		return self.tk.splitlist(
+			self.tk.call(self._w, 'window', 'names'))
+	def yview(self, *what):
+		apply(self.tk.call, (self._w, 'yview')+what)
+	def yview_pickplace(self, *what):
+		apply(self.tk.call, (self._w, 'yview', '-pickplace')+what)
+
+class OptionMenu(Widget):
+	def __init__(self, master, variable, value, *values):
+		self.widgetName = 'tk_optionMenu'
+		Widget._setup(self, master, {})
+		self.menuname = apply(
+			self.tk.call,
+			(self.widgetName, self._w, variable, value) + values)
+
+class Image:
+	def __init__(self, imgtype, name=None, cnf={}, **kw):
+		self.name = None
+		master = _default_root
+		if not master: raise RuntimeError, 'Too early to create image'
+		self.tk = master.tk
+		if not name: name = `id(self)`
+		if kw and cnf: cnf = _cnfmerge((cnf, kw))
+		elif kw: cnf = kw
+		options = ()
+		for k, v in cnf.items():
+			if type(v) in CallableTypes:
+				v = self._register(v)
+			options = options + ('-'+k, v)
+		apply(self.tk.call,
+		      ('image', 'create', imgtype, name,) + options)
+		self.name = name
+	def __str__(self): return self.name
+	def __del__(self):
+		if self.name:
+			self.tk.call('image', 'delete', self.name)
+	def __setitem__(self, key, value):
+		self.tk.call(self.name, 'configure', '-'+key, value)
+	def __getitem__(self, key):
+		return self.tk.call(self.name, 'configure', '-'+key)
+	def height(self):
+		return self.tk.getint(
+			self.tk.call('image', 'height', self.name))
+	def type(self):
+		return self.tk.call('image', 'type', self.name)
+	def width(self):
+		return self.tk.getint(
+			self.tk.call('image', 'width', self.name))
+
+class PhotoImage(Image):
+	def __init__(self, name=None, cnf={}, **kw):
+		apply(Image.__init__, (self, 'photo', name, cnf), kw)
+	def blank(self):
+		self.tk.call(self.name, 'blank')
+	def cget(self):
+		return self.tk.call(self.name, 'cget')
+	# XXX config
+	# XXX copy
+	def get(self, x, y):
+		return self.tk.call(self.name, 'get', x, y)
+	def put(self, data, to=None):
+		args = (self.name, 'put', data)
+		if to:
+			args = args + to
+		apply(self.tk.call, args)
+	# XXX read
+	# XXX write
+
+class BitmapImage(Image):
+	def __init__(self, name=None, cnf={}, **kw):
+		apply(Image.__init__, (self, 'bitmap', name, cnf), kw)
+
+def image_names(): return _default_root.tk.call('image', 'names')
+def image_types(): return _default_root.tk.call('image', 'types')
 
 ######################################################################
 # Extensions:
 
 class Studbutton(Button):
-	def __init__(self, master=None, cnf={}):
-		Widget.__init__(self, master, 'studbutton', cnf)
-		self.bind('<Any-Enter>',       self.tk_butEnter)
-		self.bind('<Any-Leave>',       self.tk_butLeave)
-		self.bind('<1>',               self.tk_butDown)
-		self.bind('<ButtonRelease-1>', self.tk_butUp)
+	def __init__(self, master=None, cnf={}, **kw):
+		Widget.__init__(self, master, 'studbutton', cnf, kw)
+		self.bind('<Any-Enter>',       self.tkButtonEnter)
+		self.bind('<Any-Leave>',       self.tkButtonLeave)
+		self.bind('<1>',               self.tkButtonDown)
+		self.bind('<ButtonRelease-1>', self.tkButtonUp)
 
 class Tributton(Button):
-	def __init__(self, master=None, cnf={}):
-		Widget.__init__(self, master, 'tributton', cnf)
-		self.bind('<Any-Enter>',       self.tk_butEnter)
-		self.bind('<Any-Leave>',       self.tk_butLeave)
-		self.bind('<1>',               self.tk_butDown)
-		self.bind('<ButtonRelease-1>', self.tk_butUp)
-
+	def __init__(self, master=None, cnf={}, **kw):
+		Widget.__init__(self, master, 'tributton', cnf, kw)
+		self.bind('<Any-Enter>',       self.tkButtonEnter)
+		self.bind('<Any-Leave>',       self.tkButtonLeave)
+		self.bind('<1>',               self.tkButtonDown)
+		self.bind('<ButtonRelease-1>', self.tkButtonUp)
+		self['fg']               = self['bg']
+		self['activebackground'] = self['bg']
