@@ -68,7 +68,7 @@ newdbmobject(file, flags, mode)
 	    else
 	        err_setstr(DbmError, (char *) gdbm_strerror(gdbm_errno));
 	    DECREF(dp);
-	    return 0;
+	    return NULL;
 	}
 	return (object *)dp;
 }
@@ -119,7 +119,7 @@ dbm_subscript(dp, key)
 	drec = gdbm_fetch(dp->di_dbm, krec);
 	if ( drec.dptr == 0 ) {
 	    err_setstr(KeyError, GETSTRINGVALUE((stringobject *)key));
-	    return 0;
+	    return NULL;
 	}
 	v = newsizedstringobject(drec.dptr, drec.dsize);
 	free(drec.dptr);
@@ -167,6 +167,20 @@ static mapping_methods dbm_as_mapping = {
 	(objobjargproc)dbm_ass_sub,	/*mp_ass_subscript*/
 };
 
+static object *
+dbm_close(dp, args)
+	register dbmobject *dp;
+	object *args;
+{
+	if ( !getnoarg(args) )
+		return NULL;
+        if ( dp->di_dbm )
+		gdbm_close(dp->di_dbm);
+	dp->di_dbm = NULL;
+	INCREF(None);
+	return None;
+}
+
 static object *
 dbm_keys(dp, args)
 	register dbmobject *dp;
@@ -174,6 +188,7 @@ dbm_keys(dp, args)
 {
 	register object *v, *item;
 	datum key, okey={ (char *)NULL, 0};
+	int err;
 
 	if (dp == NULL || !is_dbmobject(dp)) {
 		err_badcall();
@@ -187,10 +202,17 @@ dbm_keys(dp, args)
 	for (key = gdbm_firstkey(dp->di_dbm); key.dptr;
 				key = gdbm_nextkey(dp->di_dbm,okey) ) {
 	    item = newsizedstringobject(key.dptr, key.dsize);
-	    if ( item == 0 )
-	      return NULL;
-	    addlistitem(v, item);
+	    if (item == 0) {
+		    DECREF(v);
+		    return NULL;
+	    }
+	    err = addlistitem(v, item);
+	    DECREF(item);
     	    if(okey.dsize) free(okey.dptr);
+	    if (err != 0) {
+		    DECREF(v);
+		    return NULL;
+	    }
     	    okey=key;
 	}
 	return v;
@@ -271,6 +293,7 @@ dbm_reorganize(dp, args)
 }
 
 static struct methodlist dbm_methods[] = {
+	{"close",	(method)dbm_close},
 	{"keys",	(method)dbm_keys},
 	{"has_key",	(method)dbm_has_key},
 	{"firstkey",	(method)dbm_firstkey},
@@ -290,7 +313,7 @@ dbm_getattr(dp, name)
 static typeobject Dbmtype = {
 	OB_HEAD_INIT(&Typetype)
 	0,
-	"Gdbm_dictionary",
+	"gdbm",
 	sizeof(dbmobject),
 	0,
 	(destructor)dbm_dealloc, /*tp_dealloc*/
@@ -311,12 +334,14 @@ dbmopen(self, args)
     object *self;
     object *args;
 {
-    char *name, *flags;
-    int iflags, mode;
+    char *name;
+    char *flags = "r";
+    int iflags;
+   int mode = 0666;
 
 /* XXXX add other flags */
-        if ( !getargs(args, "(ssi)", &name, &flags, &mode) )
-	  return 0;
+        if ( !newgetargs(args, "s|si", &name, &flags, &mode) )
+	  return NULL;
 	if ( strcmp(flags, "r") == 0 )
 	  iflags = GDBM_READER;
 	else if ( strcmp(flags, "w") == 0 )
@@ -328,13 +353,13 @@ dbmopen(self, args)
 	else {
 	    err_setstr(DbmError,
 		       "Flags should be one of 'r', 'w', 'c' or 'n'");
-	    return 0;
+	    return NULL;
 	}
         return newdbmobject(name, iflags, mode);
 }
 
 static struct methodlist dbmmodule_methods[] = {
-    { "open", (method)dbmopen },
+    { "open", (method)dbmopen, 1 },
     { 0, 0 },
 };
 
