@@ -83,9 +83,11 @@ OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 ** IMO, mpz_get_str() assumes a bit too large target space, if he doesn't
 ** allocate it himself
 */
-#define MPZ_GET_STR_BUG
+#undef MPZ_GET_STR_BUG
 
 #include "gmp.h"
+#include "gmp-mparam.h"
+
 typedef struct {
 	OB_HEAD
         MP_INT	mpz;		/* the actual number */
