@@ -1,30 +1,7 @@
 /*
- *
- * File: timingmodule.c,v 2.2 1994/01/14 16:55:50 guido Exp
- *
  * Author: George V. Neville-Neil
- *
- * Update History: timingmodule.c,v
- * Revision 2.2  1994/01/14  16:55:50  guido
- * Make more robust against Minix and Mac
- *
- * Revision 2.1  1994/01/02  23:22:21  guido
- * Added George Neville-Neil's timing module
- *
- * Revision 1.1  93/12/28  13:14:39  gnn
- * Initial revision
- * 
- * 
- *
- *
  */
 
-#ifndef lint
-#ifndef THINK_C
-static char rcsid [] = "/ufs/guido/CVSROOT/python/Modules/timingmodule.c,v 2.2 1994/01/14 16:55:50 guido Exp" ;
-#endif /* THINK_C */
-#endif /* lint */
-
 #include "allobjects.h"
 #include "import.h"
 #include "modsupport.h"
