@@ -174,6 +174,11 @@ class_setattr(op, name, v)
 			return -1;
 		}
 	}
+	if (getrestricted()) {
+		err_setstr(RuntimeError,
+			   "classes are read-only in restricted mode");
+		return -1;
+	}
 	ac = dictlookup(op->cl_dict, name);
 	if (ac != NULL && is_accessobject(ac))
 		return setaccessvalue(ac, getowner(), v);
@@ -279,9 +284,10 @@ addaccess(class, inst)
 }
 
 object *
-newinstanceobject(class, arg)
+newinstanceobject(class, arg, kw)
 	object *class;
 	object *arg;
+	object *kw;
 {
 	register instanceobject *inst;
 	object *init;
@@ -303,16 +309,16 @@ newinstanceobject(class, arg)
 	init = instance_getattr1(inst, "__init__");
 	if (init == NULL) {
 		err_clear();
-		if (arg != NULL && !(is_tupleobject(arg) &&
-				     gettuplesize(arg) == 0)) {
+		if (arg != NULL && (!is_tupleobject(arg) ||
+				    gettuplesize(arg) != 0) || kw != NULL) {
 			err_setstr(TypeError,
-				"this classobject() takes no arguments");
+				   "this constructor takes no arguments");
 			DECREF(inst);
 			inst = NULL;
 		}
 	}
 	else {
-		object *res = call_object(init, arg);
+		object *res = PyEval_CallObjectWithKeywords(init, arg, kw);
 		DECREF(init);
 		if (res == NULL) {
 			DECREF(inst);
@@ -341,14 +347,21 @@ instance_dealloc(inst)
 	object *del;
 	/* Call the __del__ method if it exists.  First temporarily
 	   revive the object and save the current exception, if any. */
+#ifdef TRACE_REFS
+	/* much too complicated if TRACE_REFS defined */
+	extern long ref_total;
+	inst->ob_type = &Instancetype;
+	NEWREF(inst);
+	ref_total--;		/* compensate for increment in NEWREF */
+#ifdef COUNT_ALLOCS
+	inst->ob_type->tp_alloc--; /* ditto */
+#endif
+#else
 	INCREF(inst);
+#endif /* TRACE_REFS */
 	err_fetch(&error_type, &error_value, &error_traceback);
 	if ((del = instance_getattr1(inst, "__del__")) != NULL) {
-		object *args = newtupleobject(0);
-		object *res = args;
-		if (res != NULL)
-			res = call_object(del, args);
-		XDECREF(args);
+		object *res = call_object(del, (object *)NULL);
 		DECREF(del);
 		XDECREF(res);
 		/* XXX If __del__ raised an exception, it is ignored! */
@@ -356,8 +369,19 @@ instance_dealloc(inst)
 	/* Restore the saved exception and undo the temporary revival */
 	err_restore(error_type, error_value, error_traceback);
 	/* Can't use DECREF here, it would cause a recursive call */
-	if (--inst->ob_refcnt > 0)
+	if (--inst->ob_refcnt > 0) {
+#ifdef COUNT_ALLOCS
+		inst->ob_type->tp_free--;
+#endif
 		return; /* __del__ added a reference; don't delete now */
+	}
+#ifdef TRACE_REFS
+#ifdef COUNT_ALLOCS
+	inst->ob_type->tp_free--;	/* compensate for increment in UNREF */
+#endif
+	UNREF(inst);
+	inst->ob_type = NULL;
+#endif /* TRACE_REFS */
 	DECREF(inst->in_class);
 	XDECREF(inst->in_dict);
 	free((ANY *)inst);
@@ -692,7 +716,7 @@ instance_item(inst, i)
 	func = instance_getattr(inst, "__getitem__");
 	if (func == NULL)
 		return NULL;
-	arg = newintobject((long)i);
+	arg = mkvalue("(i)", i);
 	if (arg == NULL) {
 		DECREF(func);
 		return NULL;
