@@ -235,6 +235,7 @@ tok_nextc(tok)
 		else {
 			int done = 0;
 			int cur = 0;
+			char *pt;
 			if (tok->start == NULL) {
 				if (tok->buf == NULL) {
 					tok->buf = NEW(char, BUFSIZ);
@@ -271,7 +272,7 @@ tok_nextc(tok)
 					       tok->start - tok->buf;
 				int curvalid = tok->inp - tok->buf;
 				int cursize = tok->end - tok->buf;
-				int newsize = cursize + BUFSIZ;
+				int newsize = curvalid + BUFSIZ;
 				char *newbuf = tok->buf;
 				RESIZE(newbuf, char, newsize);
 				if (newbuf == NULL) {
@@ -295,6 +296,13 @@ tok_nextc(tok)
 				done = tok->inp[-1] == '\n';
 			}
 			tok->cur = tok->buf + cur;
+			/* replace "\r\n" with "\n" */
+			pt = tok->inp - 2;
+			if (pt >= tok->buf && *pt == '\r') {
+				*pt++ = '\n';
+				*pt = '\0';
+				tok->inp = pt;
+			}
 		}
 		if (tok->done != E_OK) {
 			if (tok->prompt != NULL)
@@ -416,6 +424,8 @@ tok_get(tok, p_start, p_end)
 				col++;
 			else if (c == '\t')
 				col = (col/tok->tabsize + 1) * tok->tabsize;
+			else if (c == '\014') /* Control-L (formfeed) */
+				col = 0; /* For Emacs users */
 			else
 				break;
 		}
@@ -484,7 +494,7 @@ tok_get(tok, p_start, p_end)
 	/* Skip spaces */
 	do {
 		c = tok_nextc(tok);
-	} while (c == ' ' || c == '\t');
+	} while (c == ' ' || c == '\t' || c == '\014');
 	
 	/* Set start of current token */
 	tok->start = tok->cur - 1;
