@@ -36,13 +36,13 @@ extern int verbose; /* Defined in pythonrun.c */
    symbol	-- defined for:
 
    DYNAMIC_LINK -- any kind of dynamic linking
-   USE_RLD	-- NeXT dynamic linking (currently disabled)
+   USE_RLD	-- NeXT dynamic linking
    USE_DL	-- Jack's dl for IRIX 4 or GNU dld with emulation for Jack's dl
    USE_SHLIB	-- SunOS or IRIX 5 (SVR4?) shared libraries
    _AIX		-- AIX style dynamic linking
    NT		-- NT style dynamic linking (using DLLs)
    _DL_FUNCPTR_DEFINED	-- if the typedef dl_funcptr has been defined
-   USE_MAC_SHARED_LIBRARY -- Mac CFM shared libraries
+   USE_MAC_DYNAMIC_LOADING -- Mac CFM shared libraries
    SHORT_EXT	-- short extension for dynamic module, e.g. ".so"
    LONG_EXT	-- long extension, e.g. "module.so"
    hpux		-- HP-UX Dynamic Linking - defined by the compiler
@@ -83,7 +83,7 @@ typedef FARPROC dl_funcptr;
 #define LONG_EXT "module.pyd"
 #endif
 
-#ifdef WITH_RLD
+#ifdef NeXT
 #define DYNAMIC_LINK
 #define USE_RLD
 #endif
@@ -99,10 +99,10 @@ typedef FARPROC dl_funcptr;
 #endif
 
 #ifdef __CFM68K__
-#define USE_MAC_SHARED_LIBRARY
+#define USE_MAC_DYNAMIC_LOADING
 #endif
 
-#ifdef USE_MAC_SHARED_LIBRARY
+#ifdef USE_MAC_DYNAMIC_LOADING
 #define DYNAMIC_LINK
 #define SHORT_EXT ".slb"
 #define LONG_EXT "module.slb"
@@ -127,6 +127,8 @@ static void aix_loaderror(char *name);
 #ifdef DYNAMIC_LINK
 
 #ifdef USE_SHLIB
+#include <sys/types.h>
+#include <sys/stat.h>
 #ifdef __NetBSD__
 #include <nlist.h>
 #include <link.h>
@@ -147,9 +149,9 @@ typedef void (*dl_funcptr)();
 #include "dl.h"
 #endif
 
-#ifdef USE_MAC_SHARED_LIBRARY
+#ifdef USE_MAC_DYNAMIC_LOADING
 #include <CodeFragments.h>
-#ifdef __CFM68K__ /* Really just an older version of Universal Headers */
+#ifdef SYMANTEC__CFM68K__ /* Really just an older version of Universal Headers */
 #define CFragConnectionID ConnectionID
 #define kLoadCFrag 0x01
 #endif
@@ -204,9 +206,10 @@ struct filedescr import_filetab[] = {
 };
 
 object *
-load_dynamic_module(name, pathname)
+load_dynamic_module(name, pathname, fp)
 	char *name;
 	char *pathname;
+	FILE *fp;
 {
 #ifndef DYNAMIC_LINK
 	err_setstr(ImportError, "dynamically linked modules not supported");
@@ -215,30 +218,91 @@ load_dynamic_module(name, pathname)
 	object *m;
 	char funcname[258];
 	dl_funcptr p = NULL;
+#ifdef USE_SHLIB
+	static struct {
+		dev_t dev;
+		ino_t ino;
+		void *handle;
+	} handles[128];
+	static int nhandles = 0;
+#endif
 	sprintf(funcname, FUNCNAME_PATTERN, name);
-#ifdef USE_MAC_SHARED_LIBRARY
-	/* Dynamic loading of CFM shared libraries on the Mac */
+#ifdef USE_SHLIB
+	if (fp != NULL) {
+		int i;
+		struct stat statb;
+		fstat(fileno(fp), &statb);
+		for (i = 0; i < nhandles; i++) {
+			if (statb.st_dev == handles[i].dev &&
+			    statb.st_ino == handles[i].ino) {
+				p = (dl_funcptr) dlsym(handles[i].handle,
+						       funcname);
+				goto got_it;
+			}
+		}
+		if (nhandles < 128) {
+			handles[nhandles].dev = statb.st_dev;
+			handles[nhandles].ino = statb.st_ino;
+		}
+	}
+#endif /* USE_SHLIB */
+#ifdef USE_MAC_DYNAMIC_LOADING
+	/*
+	** Dynamic loading of CFM shared libraries on the Mac.
+	** The code has become more convoluted than it was, because we want to be able
+	** to put multiple modules in a single file. For this reason, we have to determine
+	** the fragment name, and we cannot use the library entry point but we have to locate
+	** the correct init routine "by hand".
+	*/
 	{
 		FSSpec libspec;
 		CFragConnectionID connID;
 		Ptr mainAddr;
 		Str255 errMessage;
 		OSErr err;
+		Boolean isfolder, didsomething;
+		char buf[512];
+		Str63 fragname;
+		Ptr symAddr;
+		CFragSymbolClass class;
 		
+		/* First resolve any aliases to find the real file */
 		(void)FSMakeFSSpec(0, 0, Pstring(pathname), &libspec);
-		err = GetDiskFragment(&libspec, 0, 0, Pstring(name),
+#if !(defined(__MWERKS__) && defined(__CFM68K__))
+		/* Bug: not in library */
+		err = ResolveAliasFile(&libspec, 1, &isfolder, &didsomething);
+#endif
+		if ( err ) {
+			sprintf(buf, "%s: %s", pathname, PyMac_StrError(err));
+			err_setstr(ImportError, buf);
+			return NULL;
+		}
+		/* Next, determine the fragment name, by stripping '.slb' and 'module' */
+		memcpy(fragname+1, libspec.name+1, libspec.name[0]);
+		fragname[0] = libspec.name[0];
+		if( strncmp((char *)(fragname+1+fragname[0]-4), ".slb", 4) == 0 )
+			fragname[0] -= 4;
+		if ( strncmp((char *)(fragname+1+fragname[0]-6), "module", 6) == 0 )
+			fragname[0] -= 6;
+		/* Load the fragment (or return the connID if it is already loaded */
+		err = GetDiskFragment(&libspec, 0, 0, fragname, 
 				      kLoadCFrag, &connID, &mainAddr,
 				      errMessage);
 		if ( err ) {
-			char buf[512];
-			
 			sprintf(buf, "%.*s: %s", errMessage[0], errMessage+1, PyMac_StrError(err));
 			err_setstr(ImportError, buf);
 			return NULL;
 		}
-		p = (dl_funcptr)mainAddr;
+		/* Locate the address of the correct init function */
+		err = FindSymbol(connID, Pstring(funcname), &symAddr, &class);
+		if ( err ) {
+			sprintf(buf, "%s: %s", funcname, PyMac_StrError(err));
+			err_setstr(ImportError, buf);
+			return NULL;
+		}
+		p = (dl_funcptr)symAddr;
 	}
-#endif /* USE_MAC_SHARED_LIBRARY */
+#endif /* USE_MAC_DYNAMIC_LOADING */
 #ifdef USE_SHLIB
 	{
 #ifdef RTLD_NOW
@@ -255,6 +319,8 @@ load_dynamic_module(name, pathname)
 			err_setstr(ImportError, dlerror());
 			return NULL;
 		}
+		if (fp != NULL && nhandles < 128)
+			handles[nhandles++].handle = handle;
 		p = (dl_funcptr) dlsym(handle, funcname);
 	}
 #endif /* USE_SHLIB */
@@ -270,9 +336,38 @@ load_dynamic_module(name, pathname)
 		HINSTANCE hDLL;
 		hDLL = LoadLibrary(pathname);
 		if (hDLL==NULL){
-			char errBuf[64];
-			sprintf(errBuf, "DLL load failed with error code %d",
-				GetLastError());
+			char errBuf[256];
+			unsigned int errorCode;
+
+			/* Get an error string from Win32 error code */
+			char theInfo[256];           /* Pointer to error text from system */
+			int theLength;               /* Length of error text */
+
+			errorCode = GetLastError();
+
+			theLength = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, /* flags */
+				NULL,                              /* message source */
+				errorCode,                         /* the message (error) ID */
+				0,                                 /* default language environment */
+				(LPTSTR) theInfo,                  /* the buffer */
+				sizeof(theInfo),                   /* the buffer size */
+				NULL);                             /* no additional format args. */
+
+			/* Problem: could not get the error message. This should not happen if called correctly. */
+			if (theLength == 0) {
+				sprintf(errBuf, "DLL load failed with error code %d", errorCode);
+			} else {
+				int len;
+				/* For some reason a \r\n is appended to the text */
+				if (theLength >= 2 && theInfo[theLength-2] == '\r' && theInfo[theLength-1] == '\n') {
+					theLength -= 2;
+					theInfo[theLength] = '\0';
+				}
+				strcpy(errBuf, "DLL load failed: ");
+				len = strlen(errBuf);
+				strncpy(errBuf+len, theInfo, sizeof(errBuf)-len);
+				errBuf[sizeof(errBuf)-1] = '\0';
+			}
 			err_setstr(ImportError, errBuf);
 		return NULL;
 		}
@@ -345,6 +440,7 @@ load_dynamic_module(name, pathname)
                         perror(funcname);
 	}
 #endif /* hpux */
+  got_it:
 	if (p == NULL) {
 		err_setstr(ImportError,
 		   "dynamic module does not define init function");
@@ -385,7 +481,7 @@ void aix_loaderror(char *pathname)
 		int errno;
 		char *errstr;
 	} load_errtab[] = {
-		{L_ERROR_TOOMANY,	"to many errors, rest skipped."},
+		{L_ERROR_TOOMANY,	"too many errors, rest skipped."},
 		{L_ERROR_NOLIB,		"can't load library:"},
 		{L_ERROR_UNDEF,		"can't find symbol in library:"},
 		{L_ERROR_RLDBAD,
