@@ -68,7 +68,10 @@ tb_dealloc(tb)
 	DEL(tb);
 }
 
-static typeobject Tracebacktype = {
+#define Tracebacktype PyTraceBack_Type
+#define is_tracebackobject PyTraceBack_Check
+
+typeobject Tracebacktype = {
 	OB_HEAD_INIT(&Typetype)
 	0,
 	"traceback",
@@ -85,8 +88,6 @@ static typeobject Tracebacktype = {
 	0,		/*tp_as_mapping*/
 };
 
-#define is_tracebackobject(v) ((v)->ob_type == &Tracebacktype)
-
 static tracebackobject *
 newtracebackobject(next, frame, lasti, lineno)
 	tracebackobject *next;
@@ -212,7 +213,7 @@ tb_displayline(f, filename, lineno, name)
 	}
 	if (i == lineno) {
 		char *p = linebuf;
-		while (*p == ' ' || *p == '\t')
+		while (*p == ' ' || *p == '\t' || *p == '\014')
 			p++;
 		writestring("    ", f);
 		writestring(p, f);
